//Copyright 1999-2002 Gavin Conant

#include <fstream.h>
#include <iostream.h>
#include "read_seq.h"
#include "gen_dna_funcs.h"

BOOL test_interleave(char *filename)
{
  int num_taxa, num_chars;
  char dum_name[800], line[3000], test;
  ifstream infile;

  infile.open(filename);
  if(infile.fail()) {
    cerr<<"Error in Read_Sequence test_interleave: file "<<filename<<" not found\n";
    return(FALSE);
  }

  infile>>num_taxa>>num_chars;
  infile.get(dum_name, 799);
  infile.get(test);
  infile.get(line, 2999);
  infile.get(test);
  infile.get(dum_name, 799);
  infile.get(test);
  
  infile.close();

  if (test == '\t')
    return(TRUE);
  else
    return(FALSE);
  

}



Read_Sequence::Read_Sequence()
{
  in_seq=new sequence;
  in_seq_start=in_seq;
  in_seq->next_site=0;
  build_taxa= new List_Molecule_Sequence;
  taxa_start=build_taxa;
  build_taxa->next=0;
  ntaxa=0;
  size=0;
}


int Read_Sequence::get_num_taxa()
{
  return(ntaxa);
}



int Read_Sequence::get_num_chars()
{
  return(size);
}




BOOL Read_Sequence::protein_data()
{
  return(aa_data);
}


Read_Sequence::~Read_Sequence()
{
  in_seq=in_seq_start;
  build_taxa=taxa_start;
  
  while (in_seq->next_site!=0)
    {
      temp=in_seq->next_site;
      delete in_seq;
      in_seq=temp;
    }
  
  while (build_taxa->next!=0)
    {
      temp_taxa=build_taxa->next;
      delete build_taxa;
      build_taxa=temp_taxa;
    }
  delete build_taxa;
} //End Read_Sequence::~Read_Sequence

//End Read_Sequence Public functions





//Read_Sequence Private Functions
void Read_Sequence::get_sequence_array_size_known(Molecule_Sequence *new_sequence)
{ 
  int i;
  char inbase;
      
  infile.get(inbase);
        
  while(valid_input(inbase)!=TRUE)
    infile.get(inbase);
  new_sequence->Assign_site(0, convert_char(inbase));
 
  i=1;

  while((i<size) && (!infile.eof()))
    {	
      infile.get(inbase);
      if (valid_input(inbase)==TRUE)
	{
	  new_sequence->Assign_site(i, convert_char(inbase));
	  i++;
	} 
    }
  if (i<size-1)
    while (i<size) 
      {
	  new_sequence->Assign_site(i, convert_char('-'));
	  i++;
	} 
} //End Read_Sequence::get_sequence_array_size_known



void Read_Sequence::get_sequence_array(Molecule_Sequence *&new_sequence)
{
  int size_so_far=0, sitecount;
  char inbase;

  infile.get(inbase);

  in_seq=in_seq_start;

  while((valid_input(inbase)!=TRUE) && (inbase != '>'))
      infile.get(inbase);

 
  if (inbase == '>') {
    new_sequence=new Molecule_Sequence(0);
    return;
  }

  in_seq->site=convert_char(inbase);
 
  infile.get(inbase);
  size_so_far++;
  
 
  while((inbase!='>') && (inbase!='*') && (!infile.eof()))
    {	
      if (valid_input(inbase)==TRUE)
	  {
	    get_next_sequence_element(in_seq);
	    in_seq->site=convert_char(inbase);
	    size_so_far++;
	  }
      infile.get(inbase);
    }

  size=size_so_far;

  in_seq=in_seq_start;
  
  new_sequence=new Molecule_Sequence(size);

  for(sitecount=0; sitecount<size; sitecount++)
    {
      new_sequence->Assign_site(sitecount, in_seq->site);
      in_seq=in_seq->next_site;   
    }    
 
  last_char=inbase;
} //End Read_Sequence::get_sequence_array
  



void Read_Sequence::set_datatype(DATATYPE cdata)
{
  switch (cdata)
    {
    case NUCLEIC:
      aa_data=FALSE;
      valid_input=&is_base;
      convert_char=&readchar_to_base;
      break;
    case PROTEIN:
      aa_data=TRUE;
      valid_input=&is_aa;
      convert_char=&readchar_to_aa;
      break;
    }

} //End Read_Sequence::set_datatype



void Read_Sequence::get_next_list_element(List_Molecule_Sequence *&thelist)
{
  if (thelist->next==0)
    {
      thelist->next= new List_Molecule_Sequence;
      thelist->last=thelist;
      thelist->next->next=0;
    }
 
  thelist=thelist->next;
}



void Read_Sequence::get_next_sequence_element(sequence *&thelist)
{
  if (thelist->next_site==0)
    {
      in_seq->next_site=new sequence;
      in_seq->next_site->next_site=0;
    }

  in_seq=in_seq->next_site;
} //End Read_Sequence::get_next_sequence_element
//End of Read_Sequence protected functions





//Read_PIR Public functions
Sequence_dataset * Read_PIR::get_dataset(int &num_taxa, int &num_sites, char *filename, BOOL is_aa_seq)
{
  int i;
  char inbase, names[700];
  Sequence_dataset *the_data;

  if(is_aa_seq==TRUE)
    set_datatype(PROTEIN);
  else
    set_datatype(NUCLEIC);


  build_taxa= taxa_start;

  infile.open(filename);
  if (infile.fail())
    {
      cerr<<"Cannot find file "<<filename<<endl;
      return(0);
    }
  else
    {
      num_taxa=0;
      
      infile.get(inbase);
      
      while (inbase!='>')
	infile.get(inbase);
      
      while (inbase!=';')
	infile.get(inbase);
      
      infile.getline(names, 698);
      
      get_sequence_array( build_taxa->current);
    
      build_taxa->current->Assign_name(names);
      
      num_sites=size;
      get_next_list_element(build_taxa);
      
      num_taxa++;
      
      while (!infile.eof())
	{
	  infile.get(inbase);
	  
	  while (inbase!='>' && !infile.eof())
	    infile.get(inbase);
	  
	  while (inbase!=';' && !infile.eof() )
	    infile.get(inbase);
	  
	  
	  if (!infile.eof())
	    {
	      infile.getline(names, 698);
	      
	      build_taxa->current= new Molecule_Sequence(size);
  
	      get_sequence_array_size_known(build_taxa->current);
	      build_taxa->current->Assign_name(names);
	      
	      get_next_list_element(build_taxa);  
	      num_taxa++;
	    }
	}
      
      ntaxa=num_taxa;
      
      the_data= new Sequence_dataset(num_taxa, size);
      
      build_taxa=taxa_start;
      
      for(i=0; i<num_taxa; i++)
	{
	  (*the_data)[i]=*build_taxa->current;
	  
	  if (i!=num_taxa-1)
	    build_taxa=build_taxa->next;
	}
      
      infile.close();
      return(the_data);
    }
} //End Read_PIR::get_dataset



//End of Read_PIR functions




//Read_Nexus functions
Sequence_dataset * Read_Nexus::get_dataset(int &num_taxa, int &num_sites, char *filename, BOOL is_aa_seq)
{
  int i,j, namepos;
  char dump, line[202], new_name[700];
  Sequence_dataset *the_data;

  if(is_aa_seq==TRUE)
    set_datatype(PROTEIN);
  else
    set_datatype(NUCLEIC);


  infile.open(filename);
  if (infile.fail())
    {
      cerr<<"Cannot find file "<<filename<<endl;
      return(0);
    }
  else
    {
      infile.getline(line, 201);
      
      while (word_match(line, 201, "Begin Data", 10)!=1)
	infile.getline(line,201);
      
      
      while (word_match(line, 201, "dimensions", 10)!=1)
	infile.getline(line,201);
      
      
      i=loc_word_match(line, 201, "ntax", 4)+5;
      num_taxa=string_to_int(&line[i]);
      
      ntaxa=num_taxa;
      
      
      i=loc_word_match(line, 201, "nchar", 5)+6;
      num_sites=string_to_int(&line[i]);
      
      size=num_sites;
    
      the_data = new Sequence_dataset(num_taxa, size);
      
      infile.getline(line,82);
      
      while (word_match(line, 201, "Matrix", 6)!=1)
	infile.getline(line,201); 
      
      
      for ( i=0; i<num_taxa; i++)
	{
	  j=namepos=0;
	  infile>>dump;
	  while (dump!=32 && dump!='\n')
	    {       
	      if (dump!='\'')
		{
		  new_name[namepos]=dump;
		  namepos++;
		}
	      infile.get(dump);	
	      j++;
	    }
	  new_name[namepos]='\0'; 
	  
	  get_sequence_array_size_known(&(*the_data)[i]);
	  (*the_data)[i].Assign_name(new_name);
	}
      infile.close();
      return(the_data);
    }  
} //End Read_Nexus::get_dataset
//End of Read_Nexus functions
  



//Read_Phylip_interleave functions
Sequence_dataset * Read_Phylip_interleave::get_dataset(int &num_taxa, int &num_sites, char *filename, BOOL is_aa_seq)
{
  int i, j, k, l, total_count=0, end, blocks, subblocks;
  char dump, inbase, new_name[50];
  BOOL first_line=TRUE;
  Sequence_dataset *the_data;

  if(is_aa_seq==TRUE)
    set_datatype(PROTEIN);
  else
    set_datatype(NUCLEIC);

  infile.open(filename);
  if (infile.fail())
    {
      cerr<<"Cannot find file "<<filename<<endl;
      return(0);
    }
  else
    {
      
      infile>>num_taxa;
      infile>>num_sites;
      
      ntaxa=num_taxa;
      size=num_sites;
      
      the_data=new Sequence_dataset(num_taxa, size);
      
      if (size%50==0)
	blocks=(int)size/50;
      else
	blocks=(int)size/50+1;
      
      
      infile.get(dump);
      
      while(dump!='\n')
	infile.get(dump);
      
      for(l=0; l<blocks; l++)
	{
	  if (l==blocks-1)
	    if ((size-l*50)%10==0)
	      subblocks=(int)((size-l*50)/10);
	    else
	      subblocks=(int)((size-l*50)/10)+1;
	  
	  
	  else
	    subblocks=5;
	  
	  for (i=0; i<num_taxa; i++)
	    {
	      for(j=0; j<11; j++)
		if (first_line==TRUE)
	      infile.get(new_name[j]);
		else
		  infile.get(dump);
	      
	      new_name[11]='\0';

	      if (first_line == TRUE)
		(*the_data)[i].Assign_name(new_name);
	      total_count=50*(l);
	      
	      for(j=0; j<subblocks; j++)
		{

		  if (l==(blocks-1) && j == (subblocks-1))
		    end=(int)(size-((l*50)+10*(subblocks-1)));
		  else
		    end=10;
		  
		  for(k=0; k<end; k++)
		    {
		      infile.get(inbase);
		      (*the_data)[i].Assign_site(total_count, convert_char(inbase));
		      total_count++;
		    }
		  infile.get(dump);
	    }
	      while (dump!='\n')
		infile.get(dump);
	    }
	  infile.get(dump);
      first_line=FALSE;
	}
      
      infile.close();
      return(the_data);
    }
}




Sequence_dataset * Read_Phylip_noninterleave::get_dataset(int &num_taxa, int &num_sites, char *filename, BOOL is_aa_seq)
{
  int i;
  char dump, new_name[700];
  Sequence_dataset *the_data;
  
  if(is_aa_seq==TRUE)
    set_datatype(PROTEIN);
  else
    set_datatype(NUCLEIC);
  

  infile.open(filename);
  if (infile.fail())
    {
      cerr<<"Cannot find file "<<filename<<endl;
      return(0);
    }
  else
    {
      infile>>num_taxa>>num_sites;
      
      ntaxa=num_taxa;
      
      size=num_sites;
    
      the_data = new Sequence_dataset(num_taxa, size);
      
      infile.get(dump);
      
      for (i=0; i<num_taxa; i++)
	{
	 
	  infile>>new_name; 
	  
	  get_sequence_array_size_known(&(*the_data)[i]);
	  (*the_data)[i].Assign_name(new_name);
	}
      infile.close();
      return(the_data);
    }  
}


Sequence_dataset * Read_FASTA::get_dataset(int &num_taxa, int &num_sites, char *filename, BOOL is_aa_seq)
{
  int i, *sizes;
  char inbase, names[700];
  Sequence_dataset *the_data;
  
  if(is_aa_seq==TRUE)
    set_datatype(PROTEIN);
  else
    set_datatype(NUCLEIC);


  build_taxa= taxa_start;
 
  infile.clear();
  infile.open(filename);
 

  if (infile.fail())
    {
      cerr<<"Cannot find file "<<filename<<endl;
      return(0);
    }
  else
    {
      num_taxa=0;

      infile.get(inbase);
      last_char=inbase;
      while (!infile.eof())
	{
	  inbase=last_char;
	  //cout<<"Current inbase: "<<inbase<<endl; 
	  while ((inbase!='>') && !(infile.eof()))
	    infile.get(inbase);
	  if (!infile.eof())
	    {
	      infile.getline(names, 699);
	      //cout<<"Name line: "<<names<<endl;
	      get_sequence_array(build_taxa->current);
	      build_taxa->current->Assign_name(names);
	      get_next_list_element(build_taxa);
	      num_taxa++;
	    }
	 
	}
      
      ntaxa=num_taxa;
      sizes=new int [ntaxa];
      
      build_taxa=taxa_start;
      
      for (i=0; i<ntaxa; i++)
	{
	  //  cout<<"Size of sequence "<<i<<": "<<build_taxa->current->Sequence_size()<<endl;
	  sizes[i]=build_taxa->current->Sequence_size();
	  if (i!=ntaxa-1)
	    build_taxa=build_taxa->next;
	}

      the_data= new Sequence_dataset(num_taxa, sizes);
      
      build_taxa=taxa_start;
      
      for(i=0; i<num_taxa; i++)
	{
	  (*the_data)[i]=(*build_taxa->current);
	  
	  if (i!=num_taxa-1)
	    build_taxa=build_taxa->next;
	}
      
      num_sites=(*the_data)[0].Sequence_size();
      
      infile.close();
      return(the_data);
    }
}

